/*
 * AutomatonToXMLComposer.cpp
 *
 *  Created on: Nov 11, 2013
 *      Author: Jan Travnicek
 */

#include "AutomatonToXMLComposer.h"
#include "../alphabet/BlankSymbol.h"

#include "../ToXMLComposers.h"

namespace automaton {

void AutomatonToXMLComposer::composeStates(std::list<sax::Token>& out, const std::set<State>& states) const {
	out.push_back(sax::Token("states", sax::Token::TokenType::START_ELEMENT));
	for (auto& state : states) {
		out.splice(out.end(), alib::ToXMLComposers::labelComposer.compose(state.getName()));
	}
	out.push_back(sax::Token("states", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeInputAlphabet(std::list<sax::Token>& out, const std::set<alphabet::Symbol>& symbols) const {
	out.push_back(sax::Token("inputAlphabet", sax::Token::TokenType::START_ELEMENT));
	for (auto& symbol : symbols) {
		out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	}
	out.push_back(sax::Token("inputAlphabet", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeInitialStates(std::list<sax::Token>& out, const std::set<State>& states) const {
	out.push_back(sax::Token("initialStates", sax::Token::TokenType::START_ELEMENT));
	for (auto& state : states) {
		out.splice(out.end(), alib::ToXMLComposers::labelComposer.compose(state.getName()));
	}
	out.push_back(sax::Token("initialStates", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeInitialState(std::list<sax::Token>& out, const State& state) const {
	out.push_back(sax::Token("initialState", sax::Token::TokenType::START_ELEMENT));
	out.splice(out.end(), alib::ToXMLComposers::labelComposer.compose(state.getName()));
	out.push_back(sax::Token("initialState", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeFinalStates(std::list<sax::Token>& out, const std::set<State>& states) const {
	out.push_back(sax::Token("finalStates", sax::Token::TokenType::START_ELEMENT));
	for (auto& state : states) {
		out.splice(out.end(), alib::ToXMLComposers::labelComposer.compose(state.getName()));
	}
	out.push_back(sax::Token("finalStates", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeStackAlphabet(std::list<sax::Token>& out, const std::set<alphabet::Symbol>& symbols) const {
	out.push_back(sax::Token("stackAlphabet", sax::Token::TokenType::START_ELEMENT));
	for (auto& symbol : symbols) {
		out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	}
	out.push_back(sax::Token("stackAlphabet", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeInitialStackSymbols(std::list<sax::Token>& out, const std::set<alphabet::Symbol>& symbols) const {
	out.push_back(sax::Token("initialStackSymbols", sax::Token::TokenType::START_ELEMENT));
	for (auto& symbol : symbols) {
		out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	}
	out.push_back(sax::Token("initialStackSymbols", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeInitialStackSymbol(std::list<sax::Token>& out, const alphabet::Symbol& symbol) const {
	out.push_back(sax::Token("initialStackSymbol", sax::Token::TokenType::START_ELEMENT));
	out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	out.push_back(sax::Token("initialStackSymbol", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTapeAlphabet(std::list<sax::Token>& out, const std::set<alphabet::Symbol>& symbols) const {
	out.push_back(sax::Token("tapeAlphabet", sax::Token::TokenType::START_ELEMENT));
	for (auto& symbol : symbols) {
		out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	}
	out.push_back(sax::Token("tapeAlphabet", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeBlankSymbol(std::list<sax::Token>& out, const alphabet::Symbol& symbol) const {
	out.push_back(sax::Token("blankSymbol", sax::Token::TokenType::START_ELEMENT));
	out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	out.push_back(sax::Token("blankSymbol", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitions(std::list<sax::Token>& out, const UnknownAutomaton& automaton) const {
	out.push_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for (auto& transition : automaton.getTransitions()) {
		out.push_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

		if (transition.hasFrom())
			composeTransitionFrom(out, transition.getFrom());

		if(transition.hasInput())
			composeTransitionInputEpsilonSymbol(out, transition.getInput());

		if(transition.hasTo())
			composeTransitionTo(out, transition.getTo());

		if (transition.getPop().size() > 0)
			composeTransitionPop(out, transition.getPop());

		if (transition.getPush().size() > 0)
			composeTransitionPush(out, transition.getPush());

		if (transition.hasOutput())
			composeTransitionOutputEpsilonSymbol(out, transition.getOutput());

		if (transition.getShift() != Shift::NOT_SET)
			composeTransitionShift(out, transition.getShift());

		out.push_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
	}

	out.push_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}


void AutomatonToXMLComposer::composeTransitions(std::list<sax::Token>& out, const DFA& automaton) const {
	out.push_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(auto& transition : automaton.getTransitions()) {
		out.push_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

		composeTransitionFrom(out, transition.first.first);
		composeTransitionInputSymbol(out, transition.first.second);
		composeTransitionTo(out, transition.second);

		out.push_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
	}

	out.push_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitions(std::list<sax::Token>& out, const NFA& automaton) const {
	out.push_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(auto& transition : automaton.getTransitions()) {
		for(auto& targetState: transition.second) {
			out.push_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

			composeTransitionFrom(out, transition.first.first);
			composeTransitionInputSymbol(out, transition.first.second);
			composeTransitionTo(out, targetState);

			out.push_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
		}
	}

	out.push_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitions(std::list<sax::Token>& out, const EpsilonNFA& automaton) const {
	out.push_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(auto& transition : automaton.getTransitions()) {
		for(auto& targetState: transition.second) {
			out.push_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

			composeTransitionFrom(out, transition.first.first);
			composeTransitionInputEpsilonSymbol(out, transition.first.second);
			composeTransitionTo(out, targetState);

			out.push_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
		}
	}

	out.push_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitionTo(std::list<sax::Token>& out, const State& state) const {
	out.push_back(sax::Token("to", sax::Token::TokenType::START_ELEMENT));
	out.splice(out.end(), alib::ToXMLComposers::labelComposer.compose(state.getName()));
	out.push_back(sax::Token("to", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitionFrom(std::list<sax::Token>& out, const State& state) const {
	out.push_back(sax::Token("from", sax::Token::TokenType::START_ELEMENT));
	out.splice(out.end(), alib::ToXMLComposers::labelComposer.compose(state.getName()));
	out.push_back(sax::Token("from", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitionShift(std::list<sax::Token>& out, const Shift shift) const {
	out.push_back(sax::Token("shift", sax::Token::TokenType::START_ELEMENT));
	out.push_back(sax::Token(SHIFT_NAMES [shift], sax::Token::TokenType::CHARACTER));
	out.push_back(sax::Token("shift", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitionPop(std::list<sax::Token>& out, const std::vector<alphabet::Symbol>& symbols) const {
	out.push_back(sax::Token("pop", sax::Token::TokenType::START_ELEMENT));
	for (auto& symbol : symbols) {
		out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	}
	out.push_back(sax::Token("pop", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitionPush(std::list<sax::Token>& out, const std::vector<alphabet::Symbol>& symbols) const {
	out.push_back(sax::Token("push", sax::Token::TokenType::START_ELEMENT));
	for (auto& symbol : symbols) {
		out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	}
	out.push_back(sax::Token("push", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitionInputSymbol(std::list<sax::Token>& out, const alphabet::Symbol& symbol) const {
	out.push_back(sax::Token("input", sax::Token::TokenType::START_ELEMENT));
	out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	out.push_back(sax::Token("input", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitionOutputSymbol(std::list<sax::Token>& out, const alphabet::Symbol& symbol) const {
	out.push_back(sax::Token("output", sax::Token::TokenType::START_ELEMENT));
	out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol));
	out.push_back(sax::Token("output", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(std::list<sax::Token>& out, const std::variant<string::Epsilon, alphabet::Symbol>& symbol) const {
	out.push_back(sax::Token("input", sax::Token::TokenType::START_ELEMENT));
	if(symbol.is<string::Epsilon>()) {
		out.push_back(sax::Token("epsilon", sax::Token::TokenType::START_ELEMENT));
		out.push_back(sax::Token("epsilon", sax::Token::TokenType::END_ELEMENT));
	} else {
		out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol.get<alphabet::Symbol>()));
	}
	out.push_back(sax::Token("input", sax::Token::TokenType::END_ELEMENT));
}

void AutomatonToXMLComposer::composeTransitionOutputEpsilonSymbol(std::list<sax::Token>& out, const std::variant<string::Epsilon, alphabet::Symbol>& symbol) const {
	out.push_back(sax::Token("output", sax::Token::TokenType::START_ELEMENT));
	if(symbol.is<string::Epsilon>()) {
		out.push_back(sax::Token("epsilon", sax::Token::TokenType::START_ELEMENT));
		out.push_back(sax::Token("epsilon", sax::Token::TokenType::END_ELEMENT));
	} else {
		out.splice(out.end(), alib::ToXMLComposers::symbolComposer.compose(symbol.get<alphabet::Symbol>()));
	}
	out.push_back(sax::Token("output", sax::Token::TokenType::END_ELEMENT));
}

std::list<sax::Token> AutomatonToXMLComposer::compose(const AutomatonBase& automaton) const {
	std::list<sax::Token> out;
	automaton.Accept((void*) &out, *this);
	return out;
}

std::list<sax::Token> AutomatonToXMLComposer::compose(const UnknownAutomaton& automaton) const {
	std::list<sax::Token> out;
	out.push_back(sax::Token("automaton", sax::Token::TokenType::START_ELEMENT));
	
	if (automaton.getStates().size() > 0)
		composeStates(out, automaton.getStates());

	if (automaton.getInputSymbols().size() > 0)
		composeInputAlphabet(out, automaton.getInputSymbols());

	if (automaton.getTapeSymbols().size() > 0)
		composeTapeAlphabet(out, automaton.getTapeSymbols());

	if (automaton.getStackSymbols().size() > 0)
		composeStackAlphabet(out, automaton.getStackSymbols());

	if (automaton.hasBlankSymbol())
		composeBlankSymbol(out, automaton.getBlankSymbol());

	if (automaton.getInitialSymbols().size() > 0)
		composeInitialStackSymbols(out, automaton.getInitialSymbols());

	if (automaton.getInitialStates().size() > 0)
		composeInitialStates(out, automaton.getInitialStates());

	if (automaton.getFinalStates().size() > 0)
		composeFinalStates(out, automaton.getFinalStates());

	if (automaton.getTransitions().size() > 0)
		composeTransitions(out, automaton);

	out.push_back(sax::Token("automaton", sax::Token::TokenType::END_ELEMENT));
	return out;
}

std::list<sax::Token> AutomatonToXMLComposer::compose(const DFA& automaton) const {
	std::list<sax::Token> out;
	out.push_back(sax::Token("DFA", sax::Token::TokenType::START_ELEMENT));

	composeStates(out, automaton.getStates());
	composeInputAlphabet(out, automaton.getInputAlphabet());
	composeInitialState(out, automaton.getInitialState());
	composeFinalStates(out, automaton.getFinalStates());
	composeTransitions(out, automaton);

	out.push_back(sax::Token("DFA", sax::Token::TokenType::END_ELEMENT));
	return out;
}

std::list<sax::Token> AutomatonToXMLComposer::compose(const NFA& automaton) const {
	std::list<sax::Token> out;
	out.push_back(sax::Token("NFA", sax::Token::TokenType::START_ELEMENT));

	composeStates(out, automaton.getStates());
	composeInputAlphabet(out, automaton.getInputAlphabet());
	composeInitialStates(out, automaton.getInitialStates());
	composeFinalStates(out, automaton.getFinalStates());
	composeTransitions(out, automaton);

	out.push_back(sax::Token("NFA", sax::Token::TokenType::END_ELEMENT));
	return out;
}

std::list<sax::Token> AutomatonToXMLComposer::compose(const EpsilonNFA& automaton) const {
	std::list<sax::Token> out;
	out.push_back(sax::Token("EpsilonNFA", sax::Token::TokenType::START_ELEMENT));

	composeStates(out, automaton.getStates());
	composeInputAlphabet(out, automaton.getInputAlphabet());
	composeInitialStates(out, automaton.getInitialStates());
	composeFinalStates(out, automaton.getFinalStates());
	composeTransitions(out, automaton);

	out.push_back(sax::Token("EpsilonNFA", sax::Token::TokenType::END_ELEMENT));
	return out;
}

std::list<sax::Token> AutomatonToXMLComposer::compose(const ExtendedNFA& automaton) const {
	std::list<sax::Token> out;
	return out;
}

std::list<sax::Token> AutomatonToXMLComposer::compose(const CompactNFA& automaton) const {
	std::list<sax::Token> out;
	return out;
}

std::list<sax::Token> AutomatonToXMLComposer::compose(const PDA& automaton) const {
	std::list<sax::Token> out;
	return out;
}

std::list<sax::Token> AutomatonToXMLComposer::compose(const OneTapeDTM& automaton) const {
	std::list<sax::Token> out;
	return out;
}


std::list<sax::Token> AutomatonToXMLComposer::compose(const Automaton& automaton) const {
	std::list<sax::Token> out;
	automaton.getAutomaton().Accept((void*) &out, *this);
	return out;
}

void AutomatonToXMLComposer::Visit(void* data, const Automaton& automaton) const {
	*((std::list<sax::Token>*) data) = this->compose(automaton);
}

void AutomatonToXMLComposer::Visit(void* data, const UnknownAutomaton& automaton) const {
	*((std::list<sax::Token>*) data) = this->compose(automaton);
}

void AutomatonToXMLComposer::Visit(void* data, const EpsilonNFA& automaton) const {
	*((std::list<sax::Token>*) data) = this->compose(automaton);
}

void AutomatonToXMLComposer::Visit(void* data, const NFA& automaton) const {
	*((std::list<sax::Token>*) data) = this->compose(automaton);
}

void AutomatonToXMLComposer::Visit(void* data, const DFA& automaton) const {
	*((std::list<sax::Token>*) data) = this->compose(automaton);
}

void AutomatonToXMLComposer::Visit(void* data, const ExtendedNFA& automaton) const {
	*((std::list<sax::Token>*) data) = this->compose(automaton);
}

void AutomatonToXMLComposer::Visit(void* data, const CompactNFA& automaton) const {
	*((std::list<sax::Token>*) data) = this->compose(automaton);
}

void AutomatonToXMLComposer::Visit(void* data, const PDA& automaton) const {
	*((std::list<sax::Token>*) data) = this->compose(automaton);
}

void AutomatonToXMLComposer::Visit(void* data, const OneTapeDTM& automaton) const {
	*((std::list<sax::Token>*) data) = this->compose(automaton);
}

} /* namespace automaton */

