#include "FiniteAutomatonFromStringParser.h"

#include "../../exception/AlibException.h"

#include "EpsilonNFA.h"
#include "NFA.h"
#include "DFA.h"

#include "../../label/Label.h"
#include "../../label/IntegerLabel.h"

namespace automaton {

FiniteAutomatonFromStringParser::FiniteAutomatonFromStringParser(std::stringstream& input) : m_FiniteAutomatonLexer(input), m_SymbolParser(input), m_LabelParser(input) {

}

Automaton FiniteAutomatonFromStringParser::parse() {
	return parse(std::set<FEATURES>({FEATURES::EPSILON_NFA, FEATURES::NFA, FEATURES::DFA}));
}

Automaton FiniteAutomatonFromStringParser::parse(const std::set<FEATURES>& features) {
	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.token();
	if(token.type == FiniteAutomatonFromStringLexer::TokenType::EPSILON_NFA) {
		if(!features.count(FEATURES::EPSILON_NFA)) throw exception::AlibException();
		return Automaton(parseEpsilonNFA());
	} else if(token.type == FiniteAutomatonFromStringLexer::TokenType::NFA) {
		if(!features.count(FEATURES::NFA)) throw exception::AlibException();
		return Automaton(parseNFA());
	} else if(token.type == FiniteAutomatonFromStringLexer::TokenType::DFA) {
		if(!features.count(FEATURES::DFA)) throw exception::AlibException();
		return Automaton(parseDFA());
	} else {
		throw exception::AlibException();
	}
}

Automaton FiniteAutomatonFromStringParser::parseValue() {
	first();
	Automaton res = parse();

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.token();
	if(token.type == FiniteAutomatonFromStringLexer::TokenType::TEOF) {
		return std::move(res);
	} else {
		throw exception::AlibException();
	}
}

bool FiniteAutomatonFromStringParser::first() {
	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next().token();
	if(token.type == FiniteAutomatonFromStringLexer::TokenType::EPSILON_NFA || token.type == FiniteAutomatonFromStringLexer::TokenType::NFA || token.type == FiniteAutomatonFromStringLexer::TokenType::DFA) {
		return true;
	} else {
		return false;
	}
}

bool FiniteAutomatonFromStringParser::next() {
	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.next().token();
	if(token.type == FiniteAutomatonFromStringLexer::TokenType::EPSILON_NFA || token.type == FiniteAutomatonFromStringLexer::TokenType::NFA || token.type == FiniteAutomatonFromStringLexer::TokenType::DFA || token.type == FiniteAutomatonFromStringLexer::TokenType::OUT || token.type == FiniteAutomatonFromStringLexer::TokenType::IN || token.type == FiniteAutomatonFromStringLexer::TokenType::EPSILON || token.type == FiniteAutomatonFromStringLexer::TokenType::SEPARATOR || token.type == FiniteAutomatonFromStringLexer::TokenType::NONE || token.type == FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		return true;
	} else {
		return false;
	}
}

Automaton* FiniteAutomatonFromStringParser::parsePointer() {
	if(first()) {
		return new Automaton(parse());
	} else {
		return NULL;
	}
}

EpsilonNFA FiniteAutomatonFromStringParser::parseEpsilonNFA() {
	EpsilonNFA res;

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.token();
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::EPSILON_NFA) {
		throw exception::AlibException();
	}
	std::vector<std::variant<string::Epsilon, alphabet::Symbol> > symbols;

	next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
	token = m_FiniteAutomatonLexer.token();
	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		if(token.type == FiniteAutomatonFromStringLexer::TokenType::ERROR) {
			alphabet::Symbol symbol = m_SymbolParser.parse();
			res.addInputSymbol(symbol);

			std::variant<string::Epsilon, alphabet::Symbol> symbolVariant;
			symbolVariant.set<alphabet::Symbol>(symbol);
			symbols.push_back(symbolVariant);
		} else if(token.type == FiniteAutomatonFromStringLexer::TokenType::EPSILON) {
			symbols.push_back(std::variant<string::Epsilon, alphabet::Symbol>());
		}

		next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
		token = m_FiniteAutomatonLexer.token();
	}

	token = m_FiniteAutomatonLexer.token();
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) throw exception::AlibException();
	
	next() || m_LabelParser.first();
	parseEpsilonNFATransition(res, symbols);
	token = m_FiniteAutomatonLexer.token();

	while(token.type == FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		next() || m_LabelParser.first();
		token = m_FiniteAutomatonLexer.token();
		if(token.type == FiniteAutomatonFromStringLexer::TokenType::TEOF) break;

		parseEpsilonNFATransition(res, symbols);
		token = m_FiniteAutomatonLexer.token();
	}

	return  res;
}

NFA FiniteAutomatonFromStringParser::parseNFA() {
	NFA res;

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.token();
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::NFA) {
		throw exception::AlibException();
	}
	std::vector<alphabet::Symbol> symbols;

	next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
	token = m_FiniteAutomatonLexer.token();
	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		alphabet::Symbol symbol = m_SymbolParser.parse();
		res.addInputSymbol(symbol);
		symbols.push_back(symbol);

		next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
		token = m_FiniteAutomatonLexer.token();
	}

	token = m_FiniteAutomatonLexer.token();
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) throw exception::AlibException();
	
	next() || m_LabelParser.first();
	parseNFATransition(res, symbols);
	token = m_FiniteAutomatonLexer.token();

	while(token.type == FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		next() || m_LabelParser.first();
		token = m_FiniteAutomatonLexer.token();
		if(token.type == FiniteAutomatonFromStringLexer::TokenType::TEOF) break;

		parseNFATransition(res, symbols);
		token = m_FiniteAutomatonLexer.token();
	}

	return  res;
}

DFA FiniteAutomatonFromStringParser::parseDFA() {
	
	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.token();
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::DFA) {
		throw exception::AlibException();
	}
	std::vector<alphabet::Symbol> symbols;

	next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
	token = m_FiniteAutomatonLexer.token();
	while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		alphabet::Symbol symbol = m_SymbolParser.parse();
		symbols.push_back(symbol);

		next() || m_SymbolParser.first() || m_SymbolParser.m_LabelParser.first();
		token = m_FiniteAutomatonLexer.token();
	}

	token = m_FiniteAutomatonLexer.token();
	if(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) throw exception::AlibException();
	
	State* initialState = NULL;
	std::set<State> finalStates;
	std::set<State> states;
	std::set<std::tuple<State, alphabet::Symbol, State>> transitionFunction;
	
	next() || m_LabelParser.first();
	parseDFATransition(states, symbols, initialState, finalStates, transitionFunction);
	token = m_FiniteAutomatonLexer.token();

	while(token.type == FiniteAutomatonFromStringLexer::TokenType::NEW_LINE) {
		next() || m_LabelParser.first();
		token = m_FiniteAutomatonLexer.token();
		if(token.type == FiniteAutomatonFromStringLexer::TokenType::TEOF) break;

		parseDFATransition(states, symbols, initialState, finalStates, transitionFunction);
		token = m_FiniteAutomatonLexer.token();
	}

	if(initialState == NULL) throw exception::AlibException();

	DFA res(*initialState);
	delete initialState;

	res.setInputSymbols(std::set<alphabet::Symbol>(symbols.begin(), symbols.end()));
	res.setStates(states);
	res.setFinalStates(finalStates);
	for(const std::tuple<State, alphabet::Symbol, State> transition : transitionFunction) {
		res.addTransition(std::get<0>(transition), std::get<1>(transition), std::get<2>(transition));
	}

	return res;
}
	
void FiniteAutomatonFromStringParser::initialFinalState(bool& initial, bool& final) {
	initial = false;
	final = false;

	FiniteAutomatonFromStringLexer::Token token = m_FiniteAutomatonLexer.token();
	if(token.type == FiniteAutomatonFromStringLexer::TokenType::IN) {
		initial = true;
		next() || m_LabelParser.first();
		token = m_FiniteAutomatonLexer.token();

		if(token.type == FiniteAutomatonFromStringLexer::TokenType::OUT) {
			final = true;
			next() || m_LabelParser.first();
		}
	} else if(token.type == FiniteAutomatonFromStringLexer::TokenType::OUT) {
		final = true;
		next() || m_LabelParser.first();
		token = m_FiniteAutomatonLexer.token();

		if(token.type == FiniteAutomatonFromStringLexer::TokenType::IN) {
			initial = true;
			next() || m_LabelParser.first();
		}
	}
}


void FiniteAutomatonFromStringParser::parseEpsilonNFATransition(EpsilonNFA& res, const std::vector<std::variant<string::Epsilon, alphabet::Symbol> >& symbols) {
	bool initial = false;
	bool final = false;

	initialFinalState(initial, final);

	State from(m_LabelParser.parse());
	res.addState(from);
	if(initial) res.addInitialState(from);
	if(final) res.addFinalState(from);

	next() || m_LabelParser.first();

	FiniteAutomatonFromStringLexer::Token token;
	std::vector<std::variant<string::Epsilon, alphabet::Symbol>>::const_iterator iter = symbols.begin();
	do {
		if(iter == symbols.end()) throw exception::AlibException();

		token = m_FiniteAutomatonLexer.token();
		if(token.type != FiniteAutomatonFromStringLexer::TokenType::NONE) {
		
			do {
				State to(m_LabelParser.parse());
				res.addState(to);
				res.addTransition(from, *iter, to);

				next() || m_LabelParser.first();
				token = m_FiniteAutomatonLexer.token();
				if(token.type != FiniteAutomatonFromStringLexer::TokenType::SEPARATOR) break;

				next() || m_LabelParser.first();
			} while(true);
		} else {
			next() || m_LabelParser.first();
			token = m_FiniteAutomatonLexer.token();
		}
		iter++;
	} while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE);
	if(iter != symbols.end()) throw exception::AlibException();
}

void FiniteAutomatonFromStringParser::parseNFATransition(NFA& res, const std::vector<alphabet::Symbol>& symbols) {
	bool initial = false;
	bool final = false;

	initialFinalState(initial, final);

	State from(m_LabelParser.parse());
	res.addState(from);
	if(initial) res.addInitialState(from);
	if(final) res.addFinalState(from);

	next() || m_LabelParser.first();

	FiniteAutomatonFromStringLexer::Token token;
	std::vector<alphabet::Symbol>::const_iterator iter = symbols.begin();
	do {
		if(iter == symbols.end()) throw exception::AlibException();

		token = m_FiniteAutomatonLexer.token();
		if(token.type != FiniteAutomatonFromStringLexer::TokenType::NONE) {
		
			do {
				State to(m_LabelParser.parse());
				res.addState(to);
				res.addTransition(from, *iter, to);

				next() || m_LabelParser.first();
				token = m_FiniteAutomatonLexer.token();
				if(token.type != FiniteAutomatonFromStringLexer::TokenType::SEPARATOR) break;

				next() || m_LabelParser.first();
			} while(true);
		} else {
			next() || m_LabelParser.first();
			token = m_FiniteAutomatonLexer.token();
		}
		iter++;
	} while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE);
	if(iter != symbols.end()) throw exception::AlibException();
}

void FiniteAutomatonFromStringParser::parseDFATransition(std::set<State>& states, const std::vector<alphabet::Symbol>& symbols, State*& initialState, std::set<State>& finalStates, std::set<std::tuple<State, alphabet::Symbol, State>>& transitionFunction) {
	bool initial = false;
	bool final = false;

	initialFinalState(initial, final);

	State from(m_LabelParser.parse());
	states.insert(from);
	if(initial) {
		if(initialState != NULL) throw exception::AlibException();
		initialState = new State(from);
	}
	if(final) finalStates.insert(from);

	next() || m_LabelParser.first();

	FiniteAutomatonFromStringLexer::Token token;
	std::vector<alphabet::Symbol>::const_iterator iter = symbols.begin();
	do {
		if(iter == symbols.end()) throw exception::AlibException();

		token = m_FiniteAutomatonLexer.token();
		if(token.type != FiniteAutomatonFromStringLexer::TokenType::NONE) {
		
			do {
				State to(m_LabelParser.parse());
				states.insert(to);
				transitionFunction.insert(std::make_tuple(from, *iter, to));

				next() || m_LabelParser.first();
				token = m_FiniteAutomatonLexer.token();
				if(token.type != FiniteAutomatonFromStringLexer::TokenType::SEPARATOR) break;

				next() || m_LabelParser.first();
			} while(true);
		} else {
			next() || m_LabelParser.first();
			token = m_FiniteAutomatonLexer.token();
		}
		iter++;
	} while(token.type != FiniteAutomatonFromStringLexer::TokenType::NEW_LINE);
	if(iter != symbols.end()) throw exception::AlibException();
}

} /* namespace automaton */
