#include "FiniteAutomatonToStringComposer.h"
#include "../../exception/AlibException.h"

#include "../../alphabet/SymbolToStringComposer.h"
#include "../../label/LabelToStringComposer.h"

namespace automaton {

void FiniteAutomatonToStringComposer::composeTransitionsFromState(std::stringstream& out, const DFA& automaton, const State& from) const {
	std::map<std::pair<State, alphabet::Symbol>, State> symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, State>::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end()) {
			out << " -";
		} else {
			label::LabelToStringComposer composer;
			out << " " << composer.compose(toStates->second.getName());
		}
	}
}

void FiniteAutomatonToStringComposer::composeTransitionsFromState(std::stringstream& out, const NFA& automaton, const State& from) const {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > symbolTransitionsFromState = automaton.getTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const State& to : toStates->second) {
				label::LabelToStringComposer composer;
				out << (sign ? "|" : " ") << composer.compose(to.getName());
				sign = true;
			}
		}
	}
}

void FiniteAutomatonToStringComposer::composeTransitionsFromState(std::stringstream& out, const EpsilonNFA& automaton, const State& from) const {
	std::map<std::pair<State, alphabet::Symbol>, std::set<State> > symbolTransitionsFromState = automaton.getSymbolTransitionsFromState(from);

	for(const alphabet::Symbol& inputSymbol : automaton.getInputAlphabet()) {
		const std::map<std::pair<State, alphabet::Symbol>, std::set<State> >::iterator toStates = symbolTransitionsFromState.find(std::make_pair(from, inputSymbol));
		if(toStates == symbolTransitionsFromState.end() || toStates->second.size() == 0) {
			out << " -";
		} else {
			bool sign = false;
			for(const State& to : toStates->second) {
				label::LabelToStringComposer composer;
				out << (sign ? "|" : " ") << composer.compose(to.getName());
				sign = true;
			}
		}
	}

	std::map<State, std::set<State> > epsilonTransitionsFromState = automaton.getEpsilonTransitionsFromState(from);
	if(epsilonTransitionsFromState.size() == 0 || epsilonTransitionsFromState.begin()->second.size() == 0) {
		out << " -";
	} else {
		bool sign = false;
		for(const State& to : epsilonTransitionsFromState.begin()->second) {
			label::LabelToStringComposer composer;
			out << (sign ? "|" : " ") << composer.compose(to.getName());
			sign = true;
		}
	}
}

std::string FiniteAutomatonToStringComposer::compose(const DFA& automaton) const {
	std::stringstream out;

	out << "DFA";
	for(auto iterSymbol = automaton.getInputAlphabet().begin(); iterSymbol != automaton.getInputAlphabet().end(); iterSymbol++) {
		alphabet::SymbolToStringComposer composer;
		out << " " << composer.compose(*iterSymbol);
	}

	out << std::endl;

	for(auto iterState = automaton.getStates().begin(); iterState != automaton.getStates().end(); iterState++) {
		if(automaton.getInitialState() == *iterState) {
			out << ">";
		}
		if(automaton.getFinalStates().find(*iterState) != automaton.getFinalStates().end()) {
			out << "<";
		}
		label::LabelToStringComposer composer;
		out << composer.compose(iterState->getName());

		composeTransitionsFromState(out, automaton, *iterState);

		out << std::endl;
	}

	return out.str();
}

std::string FiniteAutomatonToStringComposer::compose(const NFA& automaton) const {
	std::stringstream out;

	out << "NFA";
	for(auto iterSymbol = automaton.getInputAlphabet().begin(); iterSymbol != automaton.getInputAlphabet().end(); iterSymbol++) {
		alphabet::SymbolToStringComposer composer;
		out << " " << composer.compose(*iterSymbol);
	}

	out << std::endl;

	for(auto iterState = automaton.getStates().begin(); iterState != automaton.getStates().end(); iterState++) {
		if(automaton.getInitialStates().find(*iterState) != automaton.getInitialStates().end()) {
			out << ">";
		}
		if(automaton.getFinalStates().find(*iterState) != automaton.getFinalStates().end()) {
			out << "<";
		}
		label::LabelToStringComposer composer;
		out << composer.compose(iterState->getName());

		composeTransitionsFromState(out, automaton, *iterState);

		out << std::endl;
	}

	return out.str();
}

std::string FiniteAutomatonToStringComposer::compose(const EpsilonNFA& automaton) const {
	std::stringstream out;

	out << "ENFA";
	for(auto iterSymbol = automaton.getInputAlphabet().begin(); iterSymbol != automaton.getInputAlphabet().end(); iterSymbol++) {
		alphabet::SymbolToStringComposer composer;
		out << " " << composer.compose(*iterSymbol);
	}

	out << " #E";
	out << std::endl;

	for(auto iterState = automaton.getStates().begin(); iterState != automaton.getStates().end(); iterState++) {
		if(automaton.getInitialStates().find(*iterState) != automaton.getInitialStates().end()) {
			out << ">";
		}
		if(automaton.getFinalStates().find(*iterState) != automaton.getFinalStates().end()) {
			out << "<";
		}
		label::LabelToStringComposer composer;
		out << composer.compose(iterState->getName());

		composeTransitionsFromState(out, automaton, *iterState);

		out << std::endl;
	}

	return out.str();
}

std::string FiniteAutomatonToStringComposer::compose(const Automaton& automaton) const {
	std::stringstream out;
	automaton.getAutomaton().Accept((void*) &out, *this);
	return out.str();
}

void FiniteAutomatonToStringComposer::Visit(void* data, const Automaton& automaton) const {
	*((std::stringstream*) data) << this->compose(automaton);
}

void FiniteAutomatonToStringComposer::Visit(void*, const UnknownAutomaton&) const {
	throw exception::AlibException();
}

void FiniteAutomatonToStringComposer::Visit(void* data, const EpsilonNFA& automaton) const {
	*((std::stringstream*) data) << this->compose(automaton);
}

void FiniteAutomatonToStringComposer::Visit(void* data, const NFA& automaton) const {
	*((std::stringstream*) data) << this->compose(automaton);
}

void FiniteAutomatonToStringComposer::Visit(void* data, const DFA& automaton) const {
	*((std::stringstream*) data) << this->compose(automaton);
}

void FiniteAutomatonToStringComposer::Visit(void*, const ExtendedNFA&) const {
	throw exception::AlibException();
}

void FiniteAutomatonToStringComposer::Visit(void*, const CompactNFA&) const {
	throw exception::AlibException();
}

void FiniteAutomatonToStringComposer::Visit(void*, const PDA&) const {
	throw exception::AlibException();
}

void FiniteAutomatonToStringComposer::Visit(void*, const OneTapeDTM&) const {
	throw exception::AlibException();
}

} /* namespace automaton */
