/*
 * PDA.cpp
 *
 *  Created on: Apr 10, 2013
 *      Author: martin
 */

#include "PDA.h"
#include "../../std/map.hpp"
#include "../AutomatonException.h"
#include <algorithm>

namespace automaton {

AutomatonBase* PDA::clone() const {
	return new PDA(*this);
}

AutomatonBase* PDA::plunder() && {
	return new PDA(std::move(*this));
}

bool PDA::removeState(const State& state) {
	if (initialStates.find(state) != initialStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is initial state.");
	}

	if (finalStates.find(state) != finalStates.end()) {
		throw AutomatonException("State \"" + (std::string) state.getName() + "\" is final state.");
	}

	for (std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::pair<State, std::vector<alphabet::Symbol> > > >::const_iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
		if (state == std::get<0>(transition->first))
			throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		for(std::set<std::pair<State, std::vector<alphabet::Symbol> > >::iterator target = transition->second.begin(); target != transition->second.end(); target++) {
			if(target->first == state)
				throw AutomatonException("State \"" + (std::string) state.getName() + "\" is used in transition.");
		}
	}

	return states.erase(state);
}

bool PDA::removeInputSymbol(const alphabet::Symbol& symbol) {
	for (std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::pair<State, std::vector<alphabet::Symbol> > > >::const_iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
		if (std::get<1>(transition->first).is<alphabet::Symbol>() && symbol == std::get<1>(transition->first).get<alphabet::Symbol>())
			throw AutomatonException("Symbol \"" + (std::string) symbol + "\" is used in transition.");
	}

	return inputAlphabet.erase(symbol);
}

bool PDA::addStackSymbol(const alphabet::Symbol& symbol) {
	return stackAlphabet.insert(symbol).second;
}

void PDA::setStackSymbols(const std::set<alphabet::Symbol>& newSymbols) {
	std::set<alphabet::Symbol> removed;
	std::set_difference(stackAlphabet.begin(), stackAlphabet.end(), newSymbols.begin(), newSymbols.end(), std::inserter(removed, removed.end()));

	std::set<alphabet::Symbol> added;
	std::set_difference(newSymbols.begin(), newSymbols.end(), stackAlphabet.begin(), stackAlphabet.end(), std::inserter(added, added.end()));
	
	for(const alphabet::Symbol& removedSymbol : removed) {
		removeStackSymbol(removedSymbol);
	}

	for(const alphabet::Symbol& addedSymbol : added) {
		addStackSymbol(addedSymbol);
	}
}

bool PDA::removeStackSymbol(const alphabet::Symbol& symbol) {
	for (std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::pair<State, std::vector<alphabet::Symbol> > > >::const_iterator transition = transitions.begin(); transition != transitions.end(); transition++) {
		for (std::vector<alphabet::Symbol>::const_iterator popSymbol = std::get<2>(transition->first).begin(); popSymbol != std::get<2>(transition->first).end();
				popSymbol++) {
			if (symbol == *popSymbol)
				throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
		}
		for (std::set<std::pair<State, std::vector<alphabet::Symbol> > >::const_iterator target = transition->second.begin(); target != transition->second.end(); target++) {
			if (std::find(target->second.begin(), target->second.end(), symbol) != target->second.end())
				throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is used in transition.");
		}
	}

	if(initialSymbols.find(symbol) != initialSymbols.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) symbol + "\" is start symbol.");
	}

	return stackAlphabet.erase(symbol);
}

const std::set<alphabet::Symbol>& PDA::getStackAlphabet() const {
	return stackAlphabet;
}

bool PDA::addTransition(const State& from, const alphabet::Symbol& input, const std::vector<alphabet::Symbol> pop, const State& to, const std::vector<alphabet::Symbol> push) {
	if (states.find(from) == states.end()) {
		throw AutomatonException("State \"" + (std::string) from.getName() + "\" doesn't exist.");
	}

	if (inputAlphabet.find(input) == inputAlphabet.end()) {
		throw AutomatonException("Input symbol \"" + (std::string) input + "\" doesn't exist.");
	}

	if (states.find(to) == states.end()) {
		throw AutomatonException("State \"" + (std::string) to.getName() + "\" doesn't exist.");
	}

	for(std::vector<alphabet::Symbol>::const_iterator popSymbol = pop.begin(); popSymbol != pop.end(); popSymbol++) {
		if (stackAlphabet.find(*popSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + (std::string) *popSymbol + "\" doesn't exist.");
		}
	}

	for(std::vector<alphabet::Symbol>::const_iterator pushSymbol = push.begin(); pushSymbol != push.end(); pushSymbol++) {
		if (stackAlphabet.find(*pushSymbol) == stackAlphabet.end()) {
			throw AutomatonException("Stack symbol \"" + (std::string) *pushSymbol + "\" doesn't exist.");
		}
	}

	std::variant<string::Epsilon, alphabet::Symbol> inputVariant;
	inputVariant.set<alphabet::Symbol>(input);
	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> > key(from, inputVariant, pop);
	
	std::pair<automaton::State, std::vector<alphabet::Symbol> > value = std::make_pair(to, push);
	
	return transitions[key].insert(value).second;
}

bool PDA::removeTransition(const State& from, const alphabet::Symbol& input, const std::vector<alphabet::Symbol> pop, const State& to, const std::vector<alphabet::Symbol> push) {
	std::variant<string::Epsilon, alphabet::Symbol> inputVariant;
	inputVariant.set<alphabet::Symbol>(input);
	std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> > key(from, inputVariant, pop);
	
	std::pair<automaton::State, std::vector<alphabet::Symbol> > value = std::make_pair(to, push);
	
	return transitions[key].erase(value);
}

const std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::pair<State, std::vector<alphabet::Symbol> > > >& PDA::getTransitions() const {
	return transitions;
}

bool PDA::addInitialSymbol(const alphabet::Symbol& start) {
	if (stackAlphabet.find(start) == stackAlphabet.end()) {
		throw AutomatonException("Stack symbol \"" + (std::string) start + "\" doesn't exist.");
	}
	
	return this->initialSymbols.insert(start).second;
}

bool PDA::removeInitialSymbol(const alphabet::Symbol& start) {
	return this->initialSymbols.erase(start);
}

void PDA::setInitialSymbols(const std::set<alphabet::Symbol>& symbols) {
	std::set<alphabet::Symbol> tmp;
	std::set_difference(symbols.begin(), symbols.end(), this->stackAlphabet.begin(), this->stackAlphabet.end(), std::inserter(tmp, tmp.end()));
	if(tmp.size() != 0)
		throw AutomatonException("Initial symbols not in stack alphabet");
	
	this->initialSymbols = symbols;
}

const std::set<alphabet::Symbol>& PDA::getInitialSymbols() const {
	return initialSymbols;
}

bool PDA::operator==(const AutomatonBase& other) const {
	return other == *this;
}

bool PDA::operator==(const PDA& other) const {
	return this->states == other.states && this->inputAlphabet == other.inputAlphabet && this->initialStates == other.initialStates && this->finalStates == other.finalStates && this->stackAlphabet == other.stackAlphabet && this->initialSymbols == other.initialSymbols && this->transitions == other.transitions;
}

void PDA::operator>>(std::ostream& out) const {
	out << "(PDA"
		<< "states = " << states
		<< "inputAlphabet = " << inputAlphabet
		<< "initialStates = " << initialStates
		<< "finalStates = " << finalStates
		<< "stackAlphabet = " << stackAlphabet
		<< "initialSymbols = " << initialSymbols
		<< "transitions = " << transitions
		<< ")";
}

} /* namespace automaton */
