/*
 * PDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: martin
 */

#ifndef PDA_H_
#define PDA_H_

#include <set>
#include <map>
#include <vector>
#include "../../std/variant.hpp"
#include "../AutomatonBase.h"
#include "../common/MultiInitialStates.h"
#include "../common/InputAlphabet.h"
#include "../../alphabet/Symbol.h"
#include "../../string/Epsilon.h"

namespace automaton {

/**
 * Push Down Automaton
 */
class PDA: public std::element<PDA, AutomatonBase>, public MultiInitialStates, public InputAlphabet {
protected:
	std::set<alphabet::Symbol> stackAlphabet;
	std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::pair<State, std::vector<alphabet::Symbol> > > > transitions;
	std::set<alphabet::Symbol> initialSymbols;
public:
	virtual AutomatonBase* clone() const;
	
	virtual AutomatonBase* plunder() &&;

	/**
	 * @copydoc Automaton::removeState(const State&)
	 */
	virtual bool removeState(const State& state);

	/**
	 * @copydoc Automaton::removeInputSymbol(const Symbol&)
	 */
	virtual bool removeInputSymbol(const alphabet::Symbol& symbol);

	/**
	 * Adds symbol to the stack alphabet.
	 * @param symbol Symbol to add
	 * @throws AutomatonException when symbol is already present in stack alphabet
	 */
	bool addStackSymbol(const alphabet::Symbol& symbol);

	/**
	 * Sets stack symbols of the automaton.
	 * @param symbols Symbols to set
	 * @throws AutomatonException when symbol is already present in stack alphabet
	 */
	void setStackSymbols(const std::set<alphabet::Symbol>& symbols);

	/**
	 * Removes symbol from the stack alphabet.
	 * @param symbol Symbol to remove
	 * @throws AutomatonException when symbol is used in transition
	 * (popped from or pushed to the stack), when it is a start symbol
	 * or when it is not present in the stack alphabet
	 */
	bool removeStackSymbol(const alphabet::Symbol& symbol);

	/**
	 * @return the stack alphabet
	 */
	const std::set<alphabet::Symbol>& getStackAlphabet() const;

	/**
	 * Adds transition to the PDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the PDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition(const State& from, const alphabet::Symbol& input, const std::vector<alphabet::Symbol> pop, const State& to, const std::vector<alphabet::Symbol> push);

	/**
	 * Removes the transition from the PDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the PDA
	 */
	bool removeTransition(const State& from, const alphabet::Symbol& input, const std::vector<alphabet::Symbol> pop, const State& to, const std::vector<alphabet::Symbol> push);

	/**
	 * @return PDA transitions
	 */
	const std::map<std::tuple<State, std::variant<string::Epsilon, alphabet::Symbol>, std::vector<alphabet::Symbol> >, std::set<std::pair<State, std::vector<alphabet::Symbol> > > >& getTransitions() const;

	/**
	 * Adds initial symbol. Initial symbols are symbols that are pushed
	 * to the stack when PDA is created.
	 * @param start new initial symbol
	 * @throws AutomatonException when symbol is not present in the stack alphabet or it is already in the set
	 */
	bool addInitialSymbol(const alphabet::Symbol& start);
	
	
	/**
	 * Adds initial symbol. Initial symbols are symbols that are pushed
	 * to the stack when PDA is created.
	 * @param start new initial symbol
	 * @throws AutomatonException when symbol is not present in the set of initial symbols
	 */
	bool removeInitialSymbol(const alphabet::Symbol& start);
	
	/**
	 * Sets initial symbols. Initial symbols are symbols that are pushed
	 * to the stack when PDA is created.
	 * @param symbols new initial symbols
	 * @throws AutomatonException when any of symbols is not present in the stack alphabet
	 */
	void setInitialSymbols(const std::set<alphabet::Symbol>& symbols);

	/**
	 * @return list of start symbols
	 */
	const std::set<alphabet::Symbol>& getInitialSymbols() const;

	virtual bool operator==(const AutomatonBase& other) const;

	virtual bool operator==(const PDA& other) const;

	virtual void operator>>(std::ostream& os) const;

};

} /* namespace automaton */

#endif /* PDA_H_ */
