/**
 *  https://gist.github.com/tibordp/6909880
 *  Created on: Feb 28, 2014
 *      Author: Tibor Djurica Potpara
 *      Modified: Jan Travnicek
 */

#ifndef VATIANT_HPP_
#define VATIANT_HPP_

#include <iostream>
#include <utility>
#include <typeinfo>
#include "type_traits.hpp"
#include <string>

namespace std {

template <size_t arg1, size_t ... others>
struct static_max;

template <size_t arg>
struct static_max<arg>
{
	static const size_t value = arg;
};

template <size_t arg1, size_t arg2, size_t ... others>
struct static_max<arg1, arg2, others...>
{
	static const size_t value = arg1 >= arg2 ? static_max<arg1, others...>::value :
	static_max<arg2, others...>::value;
};

template<typename... Ts>
struct variant_helper;

template<typename F, typename... Ts>
struct variant_helper<F, Ts...> {
	inline static void destroy(size_t id, void * data)
	{
		if (id == typeid(F).hash_code())
			reinterpret_cast<F*>(data)->~F();
		else
			variant_helper<Ts...>::destroy(id, data);
	}

	inline static bool move(size_t old_t, void * old_v, void * new_v)
	{
		if (old_t == typeid(F).hash_code()) {
			new (new_v) F(std::move(*reinterpret_cast<F*>(old_v)));
			return true;
		} else
			return variant_helper<Ts...>::move(old_t, old_v, new_v);
	}

	inline static void copy(size_t old_t, const void * old_v, void * new_v)
	{
		if (old_t == typeid(F).hash_code())
			new (new_v) F(*reinterpret_cast<const F*>(old_v));
		else
			return variant_helper<Ts...>::copy(old_t, old_v, new_v);
	}   
	
	inline static bool compareEq(size_t this_t, const void * this_v, size_t other_t, const void * other_v)
	{
		if (this_t == typeid(F).hash_code())
			return this_t == other_t && *(reinterpret_cast<const F*>(this_v)) == *(reinterpret_cast<const F*>(other_v));
		else
			return variant_helper<Ts...>::compareEq(this_t, this_v, other_t, other_v);
	}

	inline static void print(ostream& out, const size_t id, const void* data) {
		if (id == typeid(F).hash_code())
			out << *reinterpret_cast<const F*>(data);
		else
			variant_helper<Ts...>::print(out, id, data);
	}


	inline static bool compareLess(size_t this_t, const void * this_v, size_t other_t, const void * other_v)
	{
		if (this_t == typeid(F).hash_code() && other_t != typeid(F).hash_code()) return true;
		if (this_t != typeid(F).hash_code() && other_t == typeid(F).hash_code()) return false;

		if (this_t == typeid(F).hash_code() && other_t == typeid(F).hash_code())
			return *(reinterpret_cast<const F*>(this_v)) < *(reinterpret_cast<const F*>(other_v));
		else
			return variant_helper<Ts...>::compareLess(this_t, this_v, other_t, other_v);
	}

};

template<> struct variant_helper<>  {
inline static void destroy(size_t, void *) { }
inline static bool move(size_t, void *, void *) { return false; }
inline static void copy(size_t, const void *, void *) { }
inline static bool compareEq(size_t, const void *, size_t, const void *) { return true; }
inline static void print(ostream&, const size_t, const void *) {}
inline static bool compareLess(size_t, const void *, size_t, const void *) { return false; }
};

template<typename F, typename... Ts>
struct variant {
private:
	static const size_t data_size = static_max<sizeof(F), sizeof(Ts)...>::value;
	static const size_t data_align = static_max<alignof(F), alignof(Ts)...>::value;

	using data_t = typename std::aligned_storage<data_size, data_align>::type;

	using helper_t = variant_helper<F, Ts...>;

	size_t type_id;
	data_t data;
public: 
	//default constructor
	variant() : type_id(typeid(F).hash_code()) { new (&data) F(); }

	//copy consructor
	variant(const variant<F, Ts...>& old) : type_id(old.type_id)
	{
		helper_t::copy(old.type_id, &old.data, &data);
	}

	//move constructor
	variant(variant<F, Ts...>&& old) : type_id(old.type_id)
	{
		helper_t::move(old.type_id, &old.data, &data);
		
		new (&old.data) F();
		old.type_id = typeid(F).hash_code();
	}

	//assignment operator
	variant<F, Ts...>& operator= (variant<F, Ts...> old)
	{
		std::swap(type_id, old.type_id);
		std::swap(data, old.data);

		return *this;
	}


	bool operator== (const variant<F, Ts...>& other) const
	{
		return helper_t::compareEq(type_id, &data, other.type_id, &other.data);
	}

	bool operator!= (const variant<F, Ts...>& other) const
	{
		return !(*this == other);
	}

	bool operator< (const variant<F, Ts...>& other) const
	{
		return helper_t::compareLess(type_id, &data, other.type_id, &other.data);
	}

	bool operator> (const variant<F, Ts...>& other) const
	{
		return other < *this;
	}

	bool operator<= (const variant<F, Ts...>& other) const
	{
		return !(*this > other);
	}

	bool operator>= (const variant<F, Ts...>& other) const
	{
		return !(*this < other);
	}

	template<typename T>
	bool is() const {
		return (type_id == typeid(T).hash_code());
	}

	template<typename T>
	void set(T&& value)
	{
		if(std::is_base_of_any<T, F, Ts...>::value) {
			helper_t::destroy(type_id, &data);
			new (&data) T(value);
			type_id = typeid(T).hash_code();
		} else
			throw std::bad_cast();
	}

	template<typename T>
	void set(const T& value)
	{
		if(std::is_base_of_any<T, F, Ts...>::value) {
			helper_t::destroy(type_id, &data);
			new (&data) T(value);
			type_id = typeid(T).hash_code();
		} else
			throw std::bad_cast();
	}

	template<typename T>
	T& get()
	{
		// It is a dynamic_cast-like behaviour
		if (type_id == typeid(T).hash_code())
			return *reinterpret_cast<T*>(&data);
		else
			throw std::bad_cast();
	}

	template<typename T>
	const T& get() const
	{
		// It is a dynamic_cast-like behaviour
		if (type_id == typeid(T).hash_code())
			return *reinterpret_cast<const T*>(&data);
		else
			throw std::bad_cast();
	}

	~variant() {
		helper_t::destroy(type_id, &data);
	}

	friend std::ostream& operator <<(std::ostream& out, const variant<F, Ts...>& obj) {
		helper_t::print(out, obj.type_id, &obj.data);
		return out;
	}
};

} /* namespace std */

#endif
